# Decoder of The Geek/Furry Code
# (c) 2006 Anicka Bernathova <anicka@anicka.net> and Martin Mares <mj@ucw.cz>
# GPL'ed

use strict;
use warnings;
use lib "../lib";

package Code;

# Supplied by the definition of the code
our $code_name;
our $code_type;
our $guess_pattern;
our $intro_text;
our $bot_text;

# Parameters
our $debug = 0;
our $web = 0;
our $line = 0;
our $code_version;

# Our local subroutines
sub uncode($);
sub shell_main();
sub cgi_main();
sub out_head();
sub out_tail();

# Things we export

require Exporter;
our (@ISA, @EXPORT, @EXPORT_OK);
@ISA = ('Exporter');
@EXPORT = qw(out thunder main $debug $code_name $code_type $guess_pattern $intro_text $bot_text);
@EXPORT_OK = qw();

sub main() {
	if (defined $ENV{"GATEWAY_INTERFACE"}) {
		cgi_main();
	} else {
		shell_main();
	}
	exit 0;
}

sub shell_main() {
	require Getopt::Long;

	Getopt::Long::GetOptions(
		"debug|d!" => \$debug,
		"html|h!" => \$web,
		"line|l!" => \$line
	) or die "Usage: $0 [--debug|-d] [--html|-h] [--line|l] [<code> ...]";

	if ($web) { require UCW::CGI; }
	out_head();

	if (@ARGV) {
		foreach (@ARGV) { uncode($_); }
	} elsif ($line) {
		while (<>) {
			uncode($_);
			print "\n<", "-" x 78, ">\n\n" if !$web;
		}
	} else {
		my $code = "";
		my $open = 0;
		my $seen = 0;
		my $guess = "";
		my $guess_lines = 0;
		while (<STDIN>) {
			if (/^---+\s*BEGIN $code_type CODE BLOCK\s*---+/) {
				!$open or print STDERR "WARNING: Missing code end\n";
				$open = 1;
				$seen = 1;
				$code = "";
			} elsif (/^---+\s*END $code_type CODE BLOCK\s*---+/) {
				if ($open) {
					if ($code ne "") {
						uncode($code);
					} else {
						print STDERR "WARNING: Empty code block seen\n";
					}
					$open = 0;
					$code = "";
				} else {
					print STDERR "WARNING: Missing code begin\n";
				}
			} elsif (/^Version:\s*(\S+)/) {
				$code_version = $1;
				next;
			} elsif ($open) {
				$code .= $_;
			} elsif ($guess_lines && /^\w/) {
				$guess_lines = 0;
			} elsif (!$guess_lines && /$guess_pattern/) {
				$guess_lines = 1;
				$guess = $_;
			} elsif ($guess_lines >= 5) {
				$guess = "";
				$guess_lines = 0;
			} elsif ($guess_lines) {
				$guess .= $_;
				$guess_lines++;
			}
		}
		!$open or print STDERR "WARNING: Incomplete code block seen\n";
		if (!$seen && $guess ne "") {
			uncode($guess);
		}
	}

	out_tail();
}

sub cgi_main() {
	require UCW::CGI;

	my $code;
	my %options = (
		'code' => { 'var' => \$code, 'default' => '', 'multiline' => 1 },
	);
	UCW::CGI::parse_args(\%options);

	print "Content-type: text/html\n\n";
	$web = 1;
	out_head();

	my $c = $code;
	if ($c =~ /^Version:\s*(\S+)/) { $code_version = $1; }
	$c =~ s/^(---|Version:).*$//gm;
	$c =~ s/[ \r\n\t]+/ /g;
	$c =~ s/^ //;
	$c =~ s/ $//;
	if ($c ne "") {
		print "<p>This is what your $code_name Code spells (<a href='", UCW::CGI::self_ref('code' => $c), "'>bookmarkable link</a>):\n";
		uncode($c);
		print "<hr>\n";
	}

	my $hc = UCW::CGI::html_escape($code);
	print <<EOF ;
$intro_text

<h2>Enter Your Code:</h2>
<form action='?' method=POST>
<p><textarea name=code rows=5 cols=80>
$hc</textarea>
<p><input type=submit value=Decode>
</form>

<hr>
$bot_text
EOF

	out_tail();
}

sub thunder($) {
	my $x = shift @_;
	if ($web) {
		print "<p class=warn>WARNING: ", UCW::CGI::html_escape($x), "\n";
	} else {
		print STDERR "WARNING: $x\n";
	}
}

sub fix_punct($) {
	$_ = shift @_;
	s/([.?!])[.]+/$1/g;
	s/([.?!]),/$1 .../g;
	s/([.?!]")[.]/$1/g;
	s/[.]",/,"/g;
	s/[.]" /" /g;
	return $_;
}

my $out_line = "";
sub out($) {
	my $x = shift @_;
	$out_line .= $x;
	while ($out_line =~ /^([^\n]*)\n(.*)/s) {
		$x = $1;
		$out_line = $2;
		$x = fix_punct($x);
		if ($web) {
			if ($x =~ /^=== (.*) ===$/) {
				print "<h2>$1</h2>\n";
			} elsif ($x =~ /^== (.*) ==$/) {
				print "<h3>$1</h3>\n";
			} elsif ($x ne "") {
				print "<p>", UCW::CGI::html_escape($x), "\n";
			} else {
				print "\n";
			}
		} else {
			print "$x\n";
		}
	}
	return 1;
}

sub out_head() {
	return unless $web;
	print <<EOF ;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html40/strict.dtd">
<html><head>
<title>The $code_name Code Decoder</title>
<link rel=stylesheet title=Default href="/geek/code.css" type="text/css" media=all>
</head><body>
<h1>The $code_name Code Decoder</h1>
EOF
}

sub out_tail() {
	return unless $web;
	print <<EOF ;
</body></html>
EOF
}

sub parse($$$) {
	my ($x, $fc, $orig) = @_;
	sub unspc($) { my $x=shift @_; $x =~ tr/ /_/; return "[$x]"; }
	$x =~ s/\[([^\]]+)\]/unspc($1)/ge;
	foreach my $item (split /\s+/, $x) {
		if (my ($mod, $cat, $val) = ::parse_code($item)) {
			!defined($fc->{$cat}) or thunder "Duplicate category: $cat";
			$val =~ tr/_/ /;
			$fc->{$cat} = $mod . $val;
			$orig->{$cat} = $item;
			print "PARSE: $cat = $mod$val\n" if $debug;
		} else {
			thunder "What a devious construct: $item";
		}
	}
}

sub decode_single($$$) {
	my ($c, $val, $ctx) = @_;
	if ($c->{$val} && $val !~ /^#/) {
		out $c->{$val};
		return;
	}
	$c->{'#decode'} && &{$c->{'#decode'}} ($c, $val) and return;
	for (my $i=0; $i<=length $val; $i++) {
		my $a = substr($val, 0, $i);
		my $b = substr($val, $i);
		my ($x, $y);
		if (defined($x = $c->{"$a:"}) && defined($y = $c->{$b}) ||
		    defined($x = $c->{$a}) && defined($y = $c->{":$b"})) {
			if ($b eq "" && $ctx =~ /#/) { $y = ""; } # HACK
			out $x;
			out ", " if $x ne "" && $y ne "";
			out $y;
			return;
		}
	}
	out "Unknown value $val";
}

sub decode_simple($$$);

sub decode_simple($$$) {
	my ($c, $val, $ctx) = @_;
	if ($val eq "?" && !defined($c->{'?'})) {
		out (($code_type eq "FURRY") ? "I haven't decided yet" : "I don't know what's that.");
	} elsif ($val =~ /^(.*)\$$/) {
		decode_simple($c, $1, $ctx);
		out ", and I am getting paid for it";
	} elsif ($code_type eq "FURRY" && $val =~ /^(.*)#$/) {
		if ($1 ne "") {
			decode_simple($c, $1, "#$ctx");
			out ", and I won't tell you more";
		} else {
			out "I don't wish to tell anybody about that";
		}
	} elsif ($code_type eq "FURRY" && (my ($z) = ($val =~ /^~(.*)$/))) {
		out "Approximately ";
		decode_simple($c, $z, $ctx);
	} elsif ($code_type eq "FURRY" && (my ($x, $y) = ($val =~ /^(.+)~(.*)$/))) {
		decode_simple($c, $x, $ctx);
		out ", approximately ";
		decode_simple($c, $x . $y, $ctx);
	} elsif ($code_type eq "GEEK" && (($x, $y) = ($val =~ /^(.*)\((.*)\)$/))) {
		out "Ranging from \"";
		decode_simple($c, $x, $ctx);
		out "\" to \"";
		decode_simple($c, $y, $ctx);
		out "\"";
	} elsif ($code_type eq "GEEK" && (($x) = ($val =~ /^(.*)\@$/))) {
		out "Mostly \"";
		decode_simple($c, $x, $ctx);
		out "\", but it varies";
	} elsif ($val eq " ") {
		decode_single($c, "", $ctx);
	} else {
		decode_single($c, $val, $ctx);
	}
}

sub decode_cat($$) {
	my ($c, $val) = @_;
	if ($val =~ /^!/ && !defined($c->{'!'})) {
		out "I refuse to have anything with that!\n";
		return;
	}
	$val ne "" or $val = " ";
	my $cx = 0;
	foreach my $x (split />/, $val) {
		if ($cx && $x eq "\$") {
			out "\nBut someday I would like to get paid for it.\n";
			$cx++;
			next;
		}
		out "\nBut someday I would like to say: \"" if $cx;
		$x ne "" or $x = " ";
		my $cy = 0;
		foreach my $y (split /\//, $x) {
			out ".\n\n" if $cy;
			decode_simple($c, $y, "");
			$cy++;
		}
		out ".";
		out '"' if $cx;
		out "\n";
		$cx++;
	}
}

sub decode($$) {
	my ($fc, $orig) = @_;
	my @table = ::get_parsing_table($code_version);
	if (!@table) {
		out "== Error ==\n\n";
		out "We don't know how to parse version $code_version, sorry.\n\n";
		return;
	}
	my $cc;
	while ($cc = shift @table) {
		if ($cc =~ /^=/) { out "$cc\n\n"; next; }
		my $c = shift @table;
		my $val = $fc->{$cc};
		if (defined $val) {
			delete $fc->{$cc};
			out "== $c->{'#name'} [$orig->{$cc}] ==\n\n";
			decode_cat($c, $val);
		} else {
			out "== $c->{'#name'} ==\n\n";
			out "Nothing known.\n";
		}
		out "\n";
	}
	foreach my $c (keys %$cc) {
		thunder "Unknown category: $c";
	}
}

sub uncode($) {
	my $x = shift @_;
	my %fc = ();
	my %orig = ();
	parse($x, \%fc, \%orig);
	decode(\%fc, \%orig);
}

1;
